CREATE EXTENSION IF NOT EXISTS "pgcrypto";

-- Create tables

CREATE TABLE users (
    uuid UUID NOT NULL PRIMARY KEY,
    public_token text NOT NULL UNIQUE,
    fullname text NOT NULL,
    email text,
    admin boolean NOT NULL DEFAULT FALSE,
    created_at timestamp with time zone NOT NULL DEFAULT NOW()
);
ALTER TABLE users OWNER TO postgres;

CREATE TABLE questions (
    name text NOT NULL PRIMARY KEY,
    position integer NOT NULL,
    title text NOT NULL,
    sentence text NOT NULL,
    min_choices integer NOT NULL DEFAULT 0,
    max_choices integer DEFAULT NULL,
    vote_open boolean NOT NULL DEFAULT FALSE,
    vote_count integer NOT NULL DEFAULT 0,
    CONSTRAINT min_choices_positive CHECK (min_choices >= 0),
    CONSTRAINT nb_choices_range_valid CHECK (max_choices IS NULL OR min_choices <= max_choices)
);
ALTER TABLE questions OWNER TO postgres;

CREATE TABLE answers (
    question_name text NOT NULL REFERENCES questions(name) ON DELETE CASCADE ON UPDATE CASCADE,
    name text NOT NULL,
    position integer NOT NULL,
    checked boolean NOT NULL DEFAULT FALSE,
    sentence text NOT NULL,
    vote_count integer NOT NULL DEFAULT 0,
    PRIMARY KEY(question_name, name)
);
ALTER TABLE answers OWNER TO postgres;

CREATE TABLE votes (
    question_name text NOT NULL REFERENCES questions(name) ON DELETE CASCADE ON UPDATE CASCADE,
    user_uuid UUID NOT NULL REFERENCES users(uuid) ON DELETE RESTRICT ON UPDATE CASCADE,
    vote_time timestamp NOT NULL DEFAULT NOW(),
    PRIMARY KEY(question_name, user_uuid)
);
ALTER TABLE votes OWNER TO postgres;

-- This function is the only path to insert a new vote. The 'SECURITY DEFINER' allows this function
-- to execute with the function's privileges, like unix's setuid. That way, the database is mostly readonly,
-- and it is not possible to register votes without conforming to this function logic, which always runs
-- inside a transaction.
CREATE FUNCTION vote(arg_question_name text, arg_private_token text, arg_selected_answers text ARRAY)
RETURNS text
LANGUAGE plpgsql
SECURITY DEFINER
AS $$
DECLARE var_user_uuid UUID;
BEGIN
        -- Set var_user_uuid to user uuid from private token, or abort if the voter does not exist,
        SELECT uuid INTO var_user_uuid FROM users WHERE public_token = encode(digest(arg_private_token, 'sha256'), 'hex');
        IF var_user_uuid IS NULL
        THEN
            RETURN 'The voter is not found';
        END IF;

        -- Abort if the vote is not open, or the vote is nonexistent
        IF (SELECT COUNT(1) FROM questions WHERE name=arg_question_name AND vote_open) != 1
        THEN
            RETURN 'Vote is not open, or nonexistent';
        END IF;

        -- Abort if the number of answers is not correct
        IF array_length(arg_selected_answers, 1) < (SELECT min_choices FROM questions WHERE name=arg_question_name)
        THEN
            RETURN 'Too few answers selected in the vote';
        END IF;

        -- Abort if the number of answers is not correct
        IF array_length(arg_selected_answers, 1) > (SELECT max_choices FROM questions WHERE name=arg_question_name)
        THEN
            RETURN 'Too many answers selected in the vote';
        END IF;

        -- Abort if the answers are not in the set of valid answers
        -- this also checks that the elements in the answers set are unique
        IF array_length(arg_selected_answers, 1) != (SELECT COUNT(1) FROM answers WHERE question_name=arg_question_name AND name=ANY(arg_selected_answers))
        THEN
            RETURN 'There are duplicate or invalid answers in the vote';
        END IF;

        -- Abort if the vote already took place
        IF (SELECT COUNT(1) FROM votes WHERE question_name=arg_question_name  AND user_uuid=var_user_uuid) != 0
        THEN
            RETURN 'The vote has already taken place';
        END IF;

        -- Mark the vote so that it is not possible to vote twice
        INSERT INTO votes(question_name, user_uuid) VALUES(arg_question_name, var_user_uuid);

        -- Add 1 to the total vote count for the question
        UPDATE questions
        SET vote_count = vote_count + 1
        WHERE name=arg_question_name;

        -- Add 1 to the vote count of every selected answers
        UPDATE answers
        SET vote_count = vote_count + 1
        WHERE question_name = arg_question_name
          AND name = ANY(arg_selected_answers);

        RETURN NULL;
END;
$$;
ALTER FUNCTION vote OWNER TO postgres;


-- Drop public privileges
REVOKE ALL PRIVILEGES ON DATABASE votes                 FROM PUBLIC CASCADE;
REVOKE ALL PRIVILEGES ON SCHEMA public                  FROM PUBLIC CASCADE;
REVOKE ALL PRIVILEGES ON ALL TABLES    IN SCHEMA public FROM PUBLIC CASCADE;
REVOKE ALL PRIVILEGES ON ALL SEQUENCES IN SCHEMA public FROM PUBLIC CASCADE;
REVOKE ALL PRIVILEGES ON ALL FUNCTIONS IN SCHEMA public FROM PUBLIC CASCADE;

-- Reset votes_importer privileges: read only + can insert users,
-- but only delete user if they have not voted yet (reference constaint in place)
DROP ROLE IF EXISTS votes_importer;
CREATE ROLE votes_importer WITH LOGIN PASSWORD '438fc3dba8f85054ea7d82b332b7bf06';
REVOKE ALL PRIVILEGES ON DATABASE votes                 FROM votes_importer  CASCADE;
REVOKE ALL PRIVILEGES ON SCHEMA public                  FROM votes_importer  CASCADE;
REVOKE ALL PRIVILEGES ON ALL TABLES    IN SCHEMA public FROM votes_importer  CASCADE;
REVOKE ALL PRIVILEGES ON ALL SEQUENCES IN SCHEMA public FROM votes_importer  CASCADE;
REVOKE ALL PRIVILEGES ON ALL FUNCTIONS IN SCHEMA public FROM votes_importer  CASCADE;
GRANT CONNECT                                            ON DATABASE votes              TO votes_importer;
GRANT USAGE                                              ON SCHEMA public               TO votes_importer;
GRANT SELECT                                             ON ALL TABLES IN SCHEMA public TO votes_importer;
GRANT INSERT(uuid, public_token, fullname, email, admin) ON TABLE users                 TO votes_importer;
GRANT DELETE                                             ON TABLE users                 TO votes_importer;

-- Reset votes_app privileges: read only + can call vote() + can only update questions.vote_open column.
DROP ROLE IF EXISTS votes_app;
CREATE ROLE votes_app WITH LOGIN PASSWORD '2dd219ac1e9450b0d990025c6381ce38';
REVOKE ALL PRIVILEGES ON DATABASE votes                 FROM votes_app CASCADE;
REVOKE ALL PRIVILEGES ON SCHEMA public                  FROM votes_app CASCADE;
REVOKE ALL PRIVILEGES ON ALL TABLES    IN SCHEMA public FROM votes_app CASCADE;
REVOKE ALL PRIVILEGES ON ALL SEQUENCES IN SCHEMA public FROM votes_app CASCADE;
REVOKE ALL PRIVILEGES ON ALL FUNCTIONS IN SCHEMA public FROM votes_app CASCADE;
GRANT CONNECT            ON DATABASE votes              TO votes_app;
GRANT USAGE              ON SCHEMA public               TO votes_app;
GRANT SELECT             ON ALL TABLES IN SCHEMA public TO votes_app;
GRANT UPDATE(vote_open)  ON TABLE questions             TO votes_app;
GRANT EXECUTE            ON FUNCTION vote               TO votes_app;

-- Insert votes data
INSERT INTO questions VALUES('moral',     1, 'rapport moral',              'Rapport moral pas trop chelou ?',     1, 1);
INSERT INTO questions VALUES('financier', 2, 'rapport financier',          'Rapport financier pas trop chelou ?', 1, 1);
INSERT INTO questions VALUES('ca',        3, 'constitution du nouveau CA', 'Qui qui ya dans le CA ?',      1, NULL);

INSERT INTO answers VALUES('moral',     'oui',   1, FALSE, 'Oui !');
INSERT INTO answers VALUES('moral',     'non',   2, FALSE, 'Non');
INSERT INTO answers VALUES('moral',     'blanc', 3, FALSE, 'Blanc');

INSERT INTO answers VALUES('financier', 'oui',   1, FALSE, 'Oui !');
INSERT INTO answers VALUES('financier', 'non',   2, FALSE, 'Non');
INSERT INTO answers VALUES('financier', 'blanc', 3, FALSE, 'Blanc');

INSERT INTO answers VALUES('ca',        'a',     1, TRUE, 'Alice');
INSERT INTO answers VALUES('ca',        'b',     2, TRUE, 'Bob');
INSERT INTO answers VALUES('ca',        'c',     3, TRUE, 'Charline');
INSERT INTO answers VALUES('ca',        'd',     4, TRUE, 'Dédé');

INSERT INTO users VALUES('F066F555-E6AD-4F2E-A21F-B65933F7BF5A', '8082552deba234dfceb0fb4f019c516c476f3c2163287faa681e1c8eee0ab022', 'Robert Hacker', 'robert@example.org');
INSERT INTO users VALUES('3D023B1C-44B3-4E5C-881C-70BAFC135291', '953d2c15c76c7d2ef4867669d406c713a5e20663d57d6aa20b8ae8fd48dda309', 'Dédé', 'dd@example.org');


UPDATE questions SET vote_open=TRUE WHERE name='moral';
SELECT vote('moral', '7717e6ba488ff754fea8b57b22048c6dc6115f2c37b9fc8e87321da76045cbfb', ARRAY ['oui']);
UPDATE questions SET vote_open=FALSE WHERE name='moral';

UPDATE questions SET vote_open = TRUE WHERE name='financier';

